from pathlib import Path, PurePath
from typing import Union, Generator, Tuple, TypeVar

from ._class_file_base import FileBase
from ._helpers_private import _init_path
from ._file_name import FileName, NameABC


SelfType = TypeVar("SelfType", bound="File")


class File(FileBase):
    def __init__(self, path: Union[str, PurePath]):
        """
        :param path: to file
        """
        path = _init_path(path)
        self._path: Path = path
        self._name: FileName = FileName.from_path(path)

    def __iter__(self) -> Generator[Path, None, None]:
        """Yields 1 path"""
        yield self.path

    def __len__(self) -> int:
        """
        Always ``1``
        """
        return 1

    @property
    def name(self) -> FileName:
        """{name}{ext}"""
        return self._name

    @property
    def path(self) -> Path:
        """path to file"""
        return self._path

    def init_new(self: SelfType, path_new: Path) -> SelfType:
        return type(self)(path_new)

    def rename_iter(
        self: SelfType, name: Union[str, NameABC]
    ) -> Generator[Union[Tuple[Path, Path], SelfType], None, None]:
        """Renames file. Yields new `File` as last object"""
        new_path = self.path.with_name(str(name))
        self.path.rename(new_path)
        yield self.path, new_path
        yield type(self)(new_path)



from pathlib import Path, PurePath
from typing import Union


def _init_path(path: Union[str, PurePath]) -> Path:
    """Takes in path arg, if string, casts to Path"""
    if not isinstance(path, Path):
        path = Path(path)

    path = path.expanduser()
    return path


def _init_pure_path(path: Union[str, PurePath]) -> PurePath:
    """
    Takes in path arg, if string, casts to PurePath. Path objects are passed through,
    not cast back to PurePath.
    """
    if not isinstance(path, PurePath):
        path = PurePath(path)
    return path



import os
import shutil
from pathlib import Path
from typing import Generator, Union, Optional, cast, Tuple, TypeVar, Any

from perfsprocket import FileABC, NameABC
from perfsprocket._helpers_private import _init_path


SelfType = TypeVar("SelfType", bound="FileBase")


class FileBase(FileABC):
    def __init__(self, path: Union[str, Path], *args: Any):
        raise NotImplementedError

    def __iter__(self) -> Generator[Path, None, None]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __repr__(self) -> str:
        """<{Class Name}: '/{parent path}/{file name}'>"""
        return f"<{type(self).__name__}: '{self.path.parent / str(self.name)}'>"

    @property
    def path(self) -> Path:
        raise NotImplementedError

    @property
    def name(self) -> NameABC:
        raise NotImplementedError

    def init_new(self: SelfType, path_new: Path) -> SelfType:
        """initialize new object at end of copy or move."""
        raise NotImplementedError

    def move_iter(self: SelfType, dst_folder: Union[str, Path]) -> "BaseIterType":
        """
        Iterates through self, moving files to root level of dst_folder
        """
        dst_folder = _init_path(dst_folder)
        first: Optional[Path] = None

        for path in self:
            dst = dst_folder / path.name
            path.rename(dst)

            if first is None:
                first = dst

            yield dst

        first = cast(Path, first)
        yield self.init_new(first)

    def move(self: SelfType, dst_folder: Union[str, Path]) -> SelfType:
        """
        Executes :func:`FileBase.move_iter` and yield final item.
        """
        for item in self.move_iter(dst_folder):
            pass

        item = cast(SelfType, item)
        return item

    def copy_iter(self: SelfType, dst_folder: Union[str, Path]) -> "BaseIterType":
        """
        Iterates through self, copying files to root level of dst_folder
        """
        dst_folder = _init_path(dst_folder)
        first: Optional[Path] = None

        for path in self:
            dst = dst_folder / path.name
            shutil.copy(str(path), str(dst))

            if first is None:
                first = dst

            yield dst

        first = cast(Path, first)
        yield self.init_new(first)

    def copy(self: SelfType, dst_folder: Union[str, Path]) -> "SelfType":
        """
        Executes :func:`FileBase.copy_iter` and yields final item.
        """
        for item in self.copy_iter(dst_folder):
            pass

        item = cast(SelfType, item)
        return item

    def rename_iter(
        self: SelfType, name: Union[str, NameABC]
    ) -> Generator[Union[Tuple[Path, Path], SelfType], None, None]:
        raise NotImplementedError

    def rename(self: SelfType, name: Union[str, NameABC]) -> SelfType:
        """
        Executes :func:`FileBase.rename_iter` and yields final item.
        """
        for item in self.rename_iter(name):
            pass

        item = cast(SelfType, item)
        return item

    def delete_iter(self) -> Generator[Path, None, None]:
        """Iterates through self, deleting each path."""
        for path in self:
            os.remove(str(path))
            yield path

    def delete(self) -> None:
        """Executes :func:`delete_iter`."""
        for _ in self.delete_iter():
            pass

    def chmod_iter(self, mode: int) -> Generator[Path, None, None]:
        """Iterates through self, changing permissions on each path to ``mode``."""
        for path in self:
            path.chmod(mode)
            yield path

    def chmod(self, mode: int) -> None:
        """Executes :func:`FileBase.chmod_iter`."""
        for _ in self.chmod_iter(mode):
            pass


BaseIterType = Generator[Union[Path, SelfType], None, None]



from typing import Generator, Union, Tuple, Any, TypeVar
from typing_extensions import Protocol, runtime
from pathlib import Path

from ._file_name import NameABC


SelfType = TypeVar("SelfType", bound="FileABC")


@runtime
class FileABC(Protocol):
    def __init__(self, path: Union[str, Path], *args: Any):
        """First argument is always str or pathlib.Path object"""
        raise NotImplementedError

    def __iter__(self) -> Generator[Path, None, None]:
        """iterates through FileABC's paths, returning them"""
        raise NotImplementedError

    def __len__(self) -> int:
        """
        Number of paths contained in FileABC. Should match the number of objects yielded
        by :func:`FileABC.__iter___`
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        raise NotImplementedError

    @property
    def path(self) -> Path:
        """
        Single path most representative of file.

        :return: pathlib path. See pathlib's documentation for available methods.
        """
        raise NotImplementedError

    @property
    def name(self) -> NameABC:
        """
        :class:`NameABC` breakdown of file name.
        """
        raise NotImplementedError

    def move_iter(self: SelfType, dst_folder: Union[str, Path]) -> "FileIterType":
        """
        Yield new paths created wile moving file to ``dst_folder``. Yield new
        :class:`FileABC` at end
        """
        raise NotImplementedError

    def move(self: SelfType, dst_folder: Union[str, Path]) -> SelfType:
        """
        Non-iter version of :func:`FileABC.move_iter`, returning only new
        :class:`FileABC`
        """
        raise NotImplementedError

    def copy_iter(self: SelfType, dst_folder: Union[str, Path]) -> "FileIterType":
        """
        Yield new paths created wile moving file to ``dst_folder``. Yield new
        :class:`FileABC` at end
        """
        raise NotImplementedError

    def copy(self: SelfType, dst_folder: Union[str, Path]) -> "SelfType":
        """
        Non-iter version of :func:`FileABC.copy_iter`, returning only new
        :class:`FileABC`
        """
        raise NotImplementedError

    def rename_iter(
        self: SelfType, name: Union[str, NameABC]
    ) -> Generator[Union[Tuple[Path, Path], SelfType], None, None]:
        """
        Yields tuple of old, new paths, then new :class:`FileABC`

        :param name: Either string or :class:`NameABC` with naming details.
        """
        raise NotImplementedError

    def rename(self: SelfType, name: Union[str, NameABC]) -> SelfType:
        """
        Non-iter version of :func:`FileABC.rename_iter`, returning only new
        :class:`FileABC`
        """
        raise NotImplementedError

    def delete_iter(self) -> Generator[Path, None, None]:
        """Yields paths being deleted as operation occurs"""
        raise NotImplementedError

    def delete(self) -> None:
        """Non-iter version of :func:`FileABC.rename_iter`, returns ``None``"""
        raise NotImplementedError

    def chmod_iter(self, mode: int) -> Generator[Path, None, None]:
        """Yields paths with altered permissions as operation occurs"""
        raise NotImplementedError

    def chmod(self, mode: int) -> None:
        """Non-iter version of :func:`FileABC.chmod_iter`, returns ``None``"""
        raise NotImplementedError


FileIterType = Generator[Union[Path, SelfType], None, None]



import re
from dataclasses import dataclass
from typing import Optional, Union, Dict, Tuple, overload, Any, cast
from typing_extensions import Protocol
from pathlib import PurePath

from ._helpers_private import _init_pure_path


class Braces:
    def __init__(self, open_char: str, close_char: str):
        self._open: str = open_char
        self._close: str = close_char

    def __repr__(self) -> str:
        return f"{self.open}{self.close}"

    @property
    def open(self) -> str:
        """opening character passed to :func:`Braces.__init__`"""
        return self._open

    @property
    def close(self) -> str:
        """closing character passed to :func:`Braces.__init__`"""
        return self._close

    def enclose(self, text: str) -> str:
        """wrap text in open and close char"""
        return f"{self.open}{text}{self.close}"


ARROW = Braces("<", ">")
BRACKET = Braces("[", "]")
CURLY = Braces("{", "}")
PAREN = Braces("(", ")")


BRACES_KINDS = [ARROW, BRACKET, CURLY, PAREN]


class Flag:
    pass


KEEP = Flag()


PATTERN = re.compile(
    r"^"
    r"(?P<base>.+)"
    r"(?P<sep>[._])"
    r"(?P<open>[\[{<(])?"
    r"(?P<start>\d+|#+)"
    r"(-(?P<end>\d*|#*))?"
    r"(?P<close>[\]}>)])?"
    r"(?P<extension>\..+)?"
    r"$"
)


class NameABC(Protocol):
    base: str
    extension: Optional[str]

    def __str__(self) -> str:
        return self.formatted()

    @classmethod
    def from_path(cls, path: Union[str, PurePath]) -> "NameABC":
        """New :class:`NameABC` instance from ``pathlib.Path`` or ``str``"""
        raise NotImplementedError

    def alter(
        self,
        base: Union[str, Flag] = KEEP,
        extension: Optional[Union[str, Flag]] = KEEP,
        **kwargs: Union[Any, Flag],
    ) -> "NameABC":
        """
        Returns new `NameABC` with changes to fields. Each argument should accept the
        ``perfsprocket.KEEP`` flag as the default, which signals that a value should not
        change. ``None`` values should be signals to change a field to ``None``. Methods
        should also include **kwargs so other Name conventions arguments can be passed
        without effect.
        """
        raise NotImplementedError

    def formatted(self) -> str:
        """Returns name string."""
        raise NotImplementedError


@dataclass(frozen=True)
class FileName(NameABC):
    base: str
    extension: Optional[str]

    def __post_init__(self) -> None:
        if self.extension is not None and not self.extension.startswith("."):
            object.__setattr__(self, "extension", f".{self.extension}")

    @classmethod
    def from_path(cls, path: Union[str, PurePath]) -> "FileName":
        """New :class:`FileName` instance from ``pathlib.Path`` or ``str``"""
        path = _init_pure_path(path)
        return FileName(base=path.stem, extension=path.suffix)

    def alter(
        self,
        base: Union[str, Flag] = KEEP,
        extension: Optional[Union[str, Flag]] = KEEP,
        **kwargs: Union[Any, Flag],
    ) -> "FileName":
        """
        Make changes to name.
        ``KEEP``: keep current value
        ``None``: Change value to ``None``

        :return: new object
        """
        # We can cast this since KEEP is the only Flag type defined.
        base_arg = self.base if base is KEEP else base
        base_arg = cast(str, base_arg)

        extension_arg = self.extension if extension is KEEP else extension
        extension_arg = cast(Optional[str], extension_arg)

        return FileName(base=base_arg, extension=extension_arg)

    def formatted(self) -> str:
        """returns formatted string EX: `movie.mp4`"""
        extension = self.extension if self.extension else ""
        return f"{self.base}{extension}"


@dataclass(frozen=True)
class SeqName(FileName):
    delim: Optional[str] = "."
    start: Union[int, str] = "#"
    end: Optional[Union[int, str]] = None
    pad: int = 0
    brackets: Optional[Braces] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        object.__setattr__(self, "start", _post_init_start_end(self.start))
        object.__setattr__(self, "end", _post_init_start_end(self.end))

    @classmethod
    def from_path(cls, path: Union[str, PurePath]) -> "SeqName":
        """Parse file sequence filename from path"""
        path = _init_pure_path(path)
        pieces = _extract_pieces_from_str(path)
        base, delim, start, num_end, pad, brackets, extension = pieces

        new = cls(
            base=base,
            extension=extension,
            delim=delim,
            start=start,
            end=num_end,
            pad=pad,
            brackets=brackets,
        )
        return new

    def alter(
        self,
        base: Union[str, Flag] = KEEP,
        extension: Optional[Union[str, Flag]] = KEEP,
        delim: Union[str, Flag] = KEEP,
        start: Union[int, str, Flag] = KEEP,
        end: Optional[Union[int, str, Flag]] = KEEP,
        pad: Union[int, Flag] = KEEP,
        brackets: Optional[Union[Braces, Flag]] = KEEP,
        **kwargs: Union[Any, Flag],
    ) -> "SeqName":
        """
        Make changes to name.
        ``KEEP``: keep current value
        ``None``: Change value to ``None``

        :return: new object
        """
        new = base, extension, delim, start, end, pad, brackets
        names = ["base", "extension", "delim", "start", "end", "pad", "brackets"]

        kwargs = dict()

        for new_value, base in zip(new, names):
            if new_value is KEEP:
                new_value = getattr(self, base)
            kwargs[base] = new_value  # type: ignore

        # As long as params are typed correctly, we don't need to worry about type
        #   checks here.
        return SeqName(**kwargs)  # type: ignore

    def formatted(self) -> str:
        """returns formatted string EX: `A001_C001.[0034241-0034256].exr`"""
        open_bracket, close_bracket, range_sep = _format_range_pieces(self)
        start_str, end_str = _format_file_nums(self)
        ext = self.extension if self.extension else ""

        formatted = (
            f"{self.base}{self.delim}{open_bracket}{start_str}{range_sep}{end_str}"
            f"{close_bracket}{ext}"
        )

        return formatted


def _num_end_from_match(groups: Dict[str, Optional[str]]) -> Optional[Union[int, str]]:
    """extract end frame from REGEX match."""
    num_end: Optional[str] = groups["end"]
    if num_end:
        if "#" in num_end:
            result: Optional[Union[int, str]] = "#"
        else:
            result = int(num_end)
    else:
        result = None
    return result


def _padding_from_match(groups: Dict[str, Optional[str]]) -> int:
    """deduces padding from REGEX match"""
    start = groups["start"]
    start = cast(str, start)

    padding = len(start)

    return padding


def _brackets_from_match(groups: Dict[str, Optional[str]]) -> Optional[Braces]:
    """converts regex match to Brackets object"""
    open_char = groups["open"]
    close_char = groups["close"]

    if open_char is None and close_char is None:
        return None

    for this_bracket in BRACES_KINDS:
        if open_char != this_bracket.open:
            continue
        elif close_char != this_bracket.close:
            continue
        return this_bracket

    raise ValueError(f"Bracket Type: {open_char}{close_char} Not Recognized")


def _format_range_pieces(seq_name: "SeqName") -> Tuple[str, str, str]:
    """
    Returns open_bracket, close_bracket, range_separator. Non-relevant pieces are
    returned as empty string so concatenation does not need to run checks.
    """
    open_bracket = ""
    close_bracket = ""
    range_sep = ""

    if seq_name.end is not None:
        range_sep = "-"

        if seq_name.brackets is not None:
            open_bracket = seq_name.brackets.open
            close_bracket = seq_name.brackets.close

    return open_bracket, close_bracket, range_sep


def _format_file_nums(seq_name: "SeqName") -> Tuple[str, str]:
    fill_start = "#" if seq_name.start == "#" else "0"
    fill_end = "#" if seq_name.end == "#" else "0"

    start_str = str(seq_name.start).rjust(seq_name.pad, fill_start)

    if seq_name.end is not None:
        end_str = str(seq_name.end).rjust(seq_name.pad, fill_end)
    else:
        end_str = ""

    return start_str, end_str


def _run_filename_regex(path: PurePath) -> Dict[str, Optional[str]]:
    """extracts sequence name parts from regex"""
    file_name = path.name

    result = re.match(PATTERN, file_name)
    if result is None:
        raise ValueError

    # Stub file for re is wrong here. None can be a value if group is optional.
    #   We need to ignore MyPy complaining that Optional[str] is incorrect
    groups: Dict[str, Optional[str]] = result.groupdict()  # type: ignore
    return groups


PiecesType = Tuple[
    str,
    str,
    Union[int, str],
    Optional[Union[int, str]],
    int,
    Optional[Braces],
    Optional[str],
]


def _extract_pieces_from_str(path: PurePath) -> PiecesType:
    """extracts filename pieces from regex"""
    groups = _run_filename_regex(path)

    num_end = _num_end_from_match(groups)
    padding = _padding_from_match(groups)
    brackets = _brackets_from_match(groups)

    # need to do some casting shenanigans since our regex guarantees some returns
    base = groups["base"]
    base = cast(str, base)

    extension = groups["extension"]
    sep = groups["sep"]

    start = groups["start"]
    start = cast(str, start)

    if "#" in start:
        start_arg: Union[int, str] = "#"
    else:
        start_arg = int(start)

    sep = cast(str, sep)

    return base, sep, start_arg, num_end, padding, brackets, extension


@overload
def _post_init_start_end(value: None) -> None:
    ...


@overload  # noqa: F811
def _post_init_start_end(value: Union[int, str]) -> Union[int, str]:
    ...


def _post_init_start_end(  # noqa: F811
    value: Optional[Union[int, str]]
) -> Optional[Union[int, str]]:
    """
    parses start and end values, returns int when castable, raises value error if
    not and string is not "#"
    """
    if value is None:
        return None
    if isinstance(value, str):
        if "#" in value:
            return "#"
        else:
            raise ValueError("start/end must be int or '#'")
    else:
        return value



# "noqa" setting stops flake8 from flagging unused imports in __init__

from ._version import __version__  # noqa
from ._file_name import (
    Braces,
    BRACKET,
    PAREN,
    CURLY,
    ARROW,
    SeqName,
    FileName,
    NameABC,
    KEEP,
)
from ._file_abc import FileABC
from perfsprocket._class_file_base import FileBase
from ._class_file import File
from ._class_file_sequence import FileSequence


(
    Braces,
    BRACKET,
    PAREN,
    CURLY,
    ARROW,
    NameABC,
    FileName,
    SeqName,
    KEEP,
    FileABC,
    FileBase,
    File,
    FileSequence,
)



from pathlib import Path
from typing import Union, List, overload, Generator, Tuple, Optional, TypeVar

from perfsprocket import FileBase
from ._file_name import SeqName, NameABC, BRACKET
from ._helpers_private import _init_path


def seq_get_index(seq: "FileSequence", item: int) -> Path:
    print(str(seq.name), seq.start, seq.end, item)
    if item < 0:
        num = seq._end + item + 1
    else:
        num = seq._start + item

    if num < seq._start or num > seq.end:
        raise IndexError

    name = str(seq._name.alter(start=num, end=None))
    return seq._parent / name


def seq_get_slice(seq: "FileSequence", bounds: slice) -> "FileSequence":
    num_list: List[int] = list(range(seq._start, seq._end + 1))[bounds]
    return FileSequence(seq.path, num_list[0], num_list[-1])


@overload
def filenum_index(index: None, seq: "FileSequence") -> None:
    pass


@overload  # noqa: F811
def filenum_index(index: int, seq: "FileSequence") -> int:
    pass


def filenum_index(  # noqa: F811
    index: Optional[int], seq: "FileSequence"
) -> Optional[int]:
    """Converts file number to conventional index"""
    if index is None:
        return None
    if index < 0:
        return index
    else:
        index -= seq.start
        # if the requested index was a positive number, and adjusting for the start
        #   number throws it into the negatives, then we know the request is out of
        #   range of the sequence, so we need to raise an error here to avoid returning
        #   a file from the end of the sequence because of the negative roll over.
        if index < 0:
            raise IndexError("Frame Out of range")
        else:
            return index


class _SeqNumSlicer:
    def __init__(self, seq: "FileSequence"):
        self._seq: "FileSequence" = seq

    @overload
    def __getitem__(self, item: int) -> Path:
        ...

    @overload
    def __getitem__(self, item: slice) -> "FileSequence":  # noqa: F811
        ...

    def __getitem__(  # noqa: F811
        self, item: Union[int, slice]
    ) -> Union[Path, "FileSequence"]:
        index: Union[int, slice]

        if isinstance(item, int):
            index = filenum_index(item, self._seq)
        else:
            start = filenum_index(item.start, self._seq)
            stop = filenum_index(item.stop, self._seq)
            index = slice(start, stop, item.step)

        return self._seq[index]


SelfType = TypeVar("SelfType", bound="FileSequence")


class FileSequence(FileBase):
    def __init__(self, path: Union[str, Path], start: int, end: int):
        """
        Interact with file sequences.

        :param path: path to one (or generic) file in sequence.
            Generic path like so: 'photo_###.jpeg'
        :param start: first file number of sequence.
        :param end: last file number of sequence.
        """
        path = _init_path(path)
        name = SeqName.from_path(path).alter(start="#", end=None)

        self._parent: Path = path.parent
        self._name: SeqName = name

        self._start: int = start
        self._end: int = end

        self._seq_num_slicer: Optional["_SeqNumSlicer"] = None

    def __repr__(self) -> str:
        name = self.name.alter(start=self.start, end=self.end, brackets=BRACKET)
        return f"<{type(self).__name__}: '{self.path.parent / str(name)}'>"

    def __len__(self) -> int:
        return self._end - self._start + 1

    @overload
    def __getitem__(self, item: int) -> Path:
        ...

    @overload
    def __getitem__(self, item: slice) -> "FileSequence":  # noqa: F811
        ...

    def __getitem__(  # noqa: F811
        self, item: Union[int, slice]
    ) -> Union[Path, "FileSequence"]:
        if isinstance(item, slice):
            return seq_get_slice(self, item)
        else:
            return seq_get_index(self, item)

    def __iter__(self) -> Generator[Path, None, None]:
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self) -> Generator[Path, None, None]:
        for i in reversed(range(len(self))):
            yield self[i]

    @property
    def path(self) -> Path:
        """path to first file"""
        return self[0]

    @property
    def name(self) -> SeqName:
        """dataclass with file sequence name info. See :class:`SeqName` documentation"""
        return self._name

    @property
    def start(self) -> int:
        """Return ``start`` frame num passed to :func:`FileSequence.__init__`"""
        return self._start

    @property
    def end(self) -> int:
        """Return ``end`` frame num passed to :func:`FileSequence.__init__`"""
        return self._end

    @property
    def files(self) -> _SeqNumSlicer:
        """Returns object that can get file by frame number index."""
        if self._seq_num_slicer is None:
            self._seq_num_slicer = _SeqNumSlicer(self)
        return self._seq_num_slicer

    def init_new(self: SelfType, path_new: Path) -> SelfType:
        return type(self)(path_new, self._start, self._end)

    def rename_iter(
        self, name: Union[str, NameABC]
    ) -> Generator[Union[Tuple[Path, Path], "FileSequence"], None, None]:
        """
        Renames each file in file num order, yielding a OldPath, NewPath pair *after* it
        has been successfully moved. Yields new :class:`FileSequence` as last item.
        """
        if isinstance(name, str):
            name = SeqName.from_path(name)
        elif not isinstance(name, SeqName):
            name = SeqName(base=name.base, extension=name.extension)

        if isinstance(name.start, int):
            new_start = name.start
        else:
            new_start = self._start

        num: int
        path: Path

        first_path: Optional[Path] = None

        # this comparison is correct, since
        reverse_rename: bool = new_start < self._start

        # We need to rename form last to first if the start frame is higher. When a
        # sequence is being renamed to a higher start file, the rename will override
        # it's own files if done in order.

        new_end = new_start + len(self)
        num_range = range(new_start, new_end)
        if reverse_rename:
            path_iter = iter(self)
            num_iter = iter(num_range)
        else:
            path_iter = reversed(self)
            num_iter = reversed(num_range)

        for num, path in zip(num_iter, path_iter):
            this_name = name.alter(start=num, end=None)
            new_path = path.parent / str(this_name)
            path.rename(new_path)

            yield path, new_path

            if first_path is None:
                first_path = new_path

        if reverse_rename or first_path is None:
            first_path = new_path

        # type can be ignored here since we know all names generated with the above
        # will have an int fot this_name.start
        yield type(self)(first_path, new_start, new_end - 1)
